/**
 * PETALS - PETALS Services Platform. Copyright (c) 2006 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id: $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework.listener;

import java.net.URI;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;

import org.apache.commons.pool.impl.GenericObjectPool;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.Constants.Interceptor.Phase;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.exception.SOAP11FaultServerException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.jbidescriptor.generated.IgnoreStatus;
import org.ow2.petals.component.framework.jbidescriptor.generated.Provides;
import org.ow2.petals.component.framework.message.ExchangeImpl;
import org.ow2.petals.component.framework.performance.PerformanceNotification;
import org.ow2.petals.component.framework.su.ServiceUnitDataHandler;
import org.ow2.petals.component.framework.su.ServiceUnitManager;
import org.ow2.petals.component.framework.util.StringHelper;

/**
 * Processor for received MessageExchanges. This processor can be overridden to
 * define a specific strategy to accept messages from the DeliveryChannel. See
 * <code>acceptMessageExchange()</code>. <br>
 * Override the <code>createMessageExchangeProcessor</code> method in the
 * <code>Component</code> class to set your processor.
 * 
 * @author alouis - EBM Websourcing
 * @author ofabre - EBM Websourcing
 * @author chamerling - EBM WebSourcing
 * @since 2.0
 * 
 */
public class MessageExchangeProcessor extends Thread {

    /**
     * The component's logger
     */
    protected Logger logger;

    /**
     * The component's delivery channel
     */
    private DeliveryChannel channel;

    private AbstractComponent component;

    private Transaction currentTransaction;

    private AbstractJBIListener listener;

    private GenericObjectPool jbiPool;

    /**
     * The component's service unit manager
     */
    private ServiceUnitManager suManager;

    private ExchangeImpl exchange;

    /**
     * Creates a processor that listen to new {@link MessageExchange} on the
     * specified channel. Exchanges are ignored depending on the specified
     * exchange-status to ignore.
     * 
     * @param component
     * @param jbiPool
     * @param exchange
     */
    public MessageExchangeProcessor(AbstractComponent component, GenericObjectPool jbiPool,
            ExchangeImpl exchange, Logger logger) {
        this.component = component;
        this.jbiPool = jbiPool;
        this.channel = component.getChannel();
        this.suManager = (ServiceUnitManager) component.getServiceUnitManager();
        this.exchange = exchange;
        this.logger = logger;

        if (exchange.isTransacted()) {
            // PEtALS Container
            this.currentTransaction = (Transaction) exchange
                    .getProperty(MessageExchange.JTA_TRANSACTION_PROPERTY_NAME);

            // Potential other JBI container
            if (this.currentTransaction == null) {
                try {
                    this.currentTransaction = ((TransactionManager) this.component.getContext()
                            .getTransactionManager()).getTransaction();
                } catch (SystemException e) {
                    logger.log(Level.WARNING, "Failed to retrieve Transaction for exchange '"
                            + exchange.getExchangeId() + "'", e);
                }
            }
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Thread#run()
     */
    @Override
    public void run() {
        if (this.exchange != null) {
            if (!this.isIgnored(this.exchange)) {

                boolean exchangeTerminated = false;
                if (!this.exchange.isActiveStatus()) {
                    exchangeTerminated = true;
                }

                PerformanceNotification performanceNotification = null;
                try {
                    this.listener = (AbstractJBIListener) this.jbiPool.borrowObject();
                    performanceNotification = this.component
                            .createPerformanceNotificationReceivingFromNMR(this
                                    .getPerformanceNotificationUserData(), this.exchange);

                    this.process(this.exchange, exchangeTerminated);

                } catch (Throwable e) {
                    this.logger.log(Level.SEVERE, e.getMessage(), e);

                    // Return an error acknowledgment only of the exchange is
                    // not terminated
                    if (!exchangeTerminated) {
                        // the error may have been set by the component
                        // implementation
                        if (this.exchange.getError() == null) {
                            this.exchange.setError(new Exception(e.getMessage()));
                        }
                        try {
                            this.channel.send(this.exchange.getMessageExchange());
                        } catch (MessagingException e1) {
                            this.logger.log(Level.SEVERE,
                                    "Cannot send back an error acknowledgement", e);
                        }
                    }
                } finally {
                    this.listener.setConsumes(null);
                    this.listener.setProvides(null);
                    this.listener.setCurrentTransaction(null);
                    this.component.sendPerformanceNotification(performanceNotification);
                    try {
                        this.jbiPool.returnObject(this.listener);
                    } catch (Exception e) {
                        this.logger.warning("Can not send back listener to pool...");
                    }
                }
            }
        }
    }

    /**
     * This method should be overridden if you wish to set custom user data to
     * your performance notification.
     * 
     * @return Performance notification user data.
     */
    // TODO: Use the caller name as user data.
    protected String getPerformanceNotificationUserData() {
        return "";
    }

    /**
     * Launch the component processing of the given message exchange.
     * 
     * @param exchange
     *            The message exchange to process. Must be non null
     * @param exchangeTerminated
     *            {@code True} is the exchange is terminated
     * @throws InterruptedException
     * @throws MessagingException
     *             A problem occurred during manipulation of the exchange
     */
    protected void process(final Exchange exchange, boolean exchangeTerminated)
            throws MessagingException, InterruptedException {

        if (exchange.isProviderRole()) {
            this.logger.finest("Process an exchange as PROVIDER with id : "
                    + exchange.getExchangeId());
            Provides provides = this.getProvidesFromExchange(exchange);
            if (provides == null) {
                throw new MessagingException("The endpoint '" + exchange.getEndpoint()
                        + "' is unknown on the component : "
                        + this.component.getComponentConfiguration().getIdentification().getName());
            }

            ConfigurationExtensions extensions = this.getExtensions(provides);

            this.listener.setProvides(provides);
            this.listener.setExtensions(extensions);
            this.listener.setCurrentTransaction(this.currentTransaction);

            try {
                this.listener.getMessageInterceptorProcessor().intercept(exchange, provides,
                        Phase.ACCEPT);
            } catch (PEtALSCDKException e) {
                throw new MessagingException(e);
            }

            this.processAsProvider(exchange, exchangeTerminated);

            try {
                this.listener.getMessageInterceptorProcessor().intercept(exchange, provides,
                        Phase.ACCEPT_RESPONSE);
            } catch (PEtALSCDKException e) {
                throw new MessagingException(e);
            }

            if (!exchangeTerminated) {
                if (provides.getExchangeProperties() != null) {
                    ConfigurationExtensions configurationExtensions = null;
                    if (this.component.getComponentConfiguration().getPropertiesFile() != null
                            && !StringHelper.isNullOrEmpty(this.component
                                    .getComponentConfiguration().getPropertiesFile().getValue())) {
                        try {
                            configurationExtensions = new ConfigurationExtensions(provides
                                    .getExchangeProperties().getExchangeProperty(), this.component
                                    .getComponentConfiguration().getPropertiesFile().getValue());
                        } catch (PEtALSCDKException e) {
                            this.logger
                                    .warning("Failed to get configured Exchange Properties against component properties"
                                            + e.getMessage());
                        }
                        if (configurationExtensions == null) {
                            configurationExtensions = new ConfigurationExtensions(provides
                                    .getExchangeProperties().getExchangeProperty());
                        }
                        for (String name : configurationExtensions.keySet()) {
                            exchange.setProperty(name, configurationExtensions.get(name));
                        }
                    }
                }
                this.channel.send(((ExchangeImpl) exchange).getMessageExchange());
            }

        } else {
            this.logger.finest("Process an exchange as CONSUMER with id : "
                    + exchange.getExchangeId());

            Consumes consumes = this.getConsumesFromExchange(exchange);
            if (consumes == null) {
                throw new MessagingException("The exchange '" + exchange.getExchangeId()
                        + "' refers to an unknown consumer on the component : "
                        + this.component.getComponentConfiguration().getIdentification().getName());
            }

            try {
                this.listener.getMessageInterceptorProcessor().intercept(exchange, consumes,
                        Phase.SEND_RESPONSE);
            } catch (PEtALSCDKException e) {
                throw new MessagingException(e);
            }

            // Notify a synchronizer if exist
            MessageExchangeSynchronizer exchangeSynchronizer = this.component
                    .getAsyncMessageHandler().getAsynchronousMessage(exchange.getExchangeId());
            if (exchangeSynchronizer != null) {
                if (!exchangeSynchronizer.isTimeout()) {
                    exchangeSynchronizer.setExchange(exchange);
                    exchangeSynchronizer.notifyExchange();
                } else if (!exchangeTerminated) {
                    exchange
                            .setError(new Exception(
                                    "The response to an asynchronous exchange "
                                            + "arrived too late, the corellation with the request is not possible"));
                    this.channel.send(((ExchangeImpl) exchange).getMessageExchange());
                }
            } else {
                this.listener.setConsumes(consumes);
                this.listener.setExtensions(this.getExtensions(consumes));
                this.listener.setCurrentTransaction(this.currentTransaction);

                this.processAsConsumer(exchange);

                if (!exchangeTerminated) {
                    this.channel.send(((ExchangeImpl) exchange).getMessageExchange());
                }
            }
        }
    }

    /**
     * Depending on the IgnoredStatus, tells if the exchange has to be ignored.
     * If the exchange is null, return true.
     * 
     * @param exchange
     * @return
     */
    protected boolean isIgnored(final Exchange exchange) {
        boolean result = false;

        if (((ExchangeImpl) exchange).getMessageExchange() == null) {
            result = true;
        } else {
            IgnoreStatus status = this.component.getComponentConfiguration().getIgnoredStatus();
            if (exchange.isDoneStatus()) {
                result = IgnoreStatus.DONE_IGNORED == status
                        || IgnoreStatus.DONE_AND_ERROR_IGNORED == status;
            } else if (exchange.isErrorStatus()) {
                result = IgnoreStatus.ERROR_IGNORED == status
                        || IgnoreStatus.DONE_AND_ERROR_IGNORED == status;
            }
        }
        return result;
    }

    /**
     * If the received exchange is a fault response from the consumer,
     * acknowledge it : status DONE
     * 
     * @param exchange
     *            The processed {@link MessageExchange}
     * @return true if the ack. of the Fault is done; false if no fault is
     *         present in the exchange
     * @throws MessagingException
     */
    private boolean ackFaultReception(final Exchange exchange) throws MessagingException {
        boolean result = false;

        if (exchange.getFault() != null) {
            this.listener.onJBIMessage(exchange);
            exchange.setDoneStatus();
            result = true;
        }
        return result;
    }

    /**
     * Get the Consumes element for the given exchange.
     * 
     * @param exchange
     * @return {@code Consumes}
     */
    protected Consumes getConsumesFromExchange(Exchange exchange) {
        ServiceEndpoint endpoint = exchange.getEndpoint();
        return this.suManager.getConsumesFromDestination(endpoint.getEndpointName(), exchange
                .getService(), exchange.getInterfaceName());
    }

    /**
     * Get the Provides element for the given exchange
     * 
     * @param exchange
     * @return {@code Provides}
     */
    protected Provides getProvidesFromExchange(final Exchange exchange) {
        ServiceEndpoint ep = exchange.getEndpoint();
        Provides provides = this.suManager.getProvidesFromEndpoint(ep);
        return provides;
    }

    /**
     * Get the extensions for the given provides/consumes.
     * 
     * @param service
     * @return The extensions
     */
    private ConfigurationExtensions getExtensions(final Object service) {
        ConfigurationExtensions extensions = null;

        ServiceUnitDataHandler suDataHandler = this.suManager.getSUDataHandlerForService(service);
        if (suDataHandler != null) {
            extensions = suDataHandler.getConfigurationExtensions(service);
        }

        return extensions;
    }

    /**
     * Process the exchange as service consumer
     * 
     * @param exchange
     * @throws MessagingException
     */
    private void processAsConsumer(final Exchange exchange) throws MessagingException {
        if (exchange.isRobustInOnlyPattern()) {
            this.processRobustInOnlyAsConsumer(exchange);
        } else if (exchange.isInOnlyPattern()) {
            this.processInOnlyAsConsumer(exchange);
        } else if (exchange.isInOutPattern()) {
            this.processInOutAsConsumer(exchange);
        } else if (exchange.isInOptionalOutPattern()) {
            this.processInOptionalOutAsConsumer(exchange);
        } else {
            throw new MessagingException("MessageExchangePattern not recognized :"
                    + exchange.getPattern());
        }
    }

    /**
     * Process the exchange as service provider
     * 
     * @param exchange
     *            The ongoing exchange
     * @param exchange
     *            {@code True} if the exchange is terminated
     * @throws MessagingException
     */
    private void processAsProvider(final Exchange exchange, boolean exchangeTerminated)
            throws MessagingException {
        if (exchange.isInOnlyPattern()) {
            this.processInOnlyAsProvider(exchange);
        } else if (exchange.isRobustInOnlyPattern()) {
            this.processRobustInOnlyAsProvider(exchange);
        } else if (exchange.isInOutPattern()) {
            this.processInOutAsProvider(exchange, exchangeTerminated);
        } else if (exchange.isInOptionalOutPattern()) {
            this.processInOptionalOutAsProvider(exchange);
        } else {
            throw new MessagingException("MessageExchangePattern not recognized :"
                    + exchange.getPattern());
        }
    }

    /**
     * 
     * @param exchange
     * @throws MessagingException
     */
    private void processInOnlyAsConsumer(Exchange exchange) throws MessagingException {
        if (exchange.isDoneStatus() || exchange.isErrorStatus()) {
            this.listener.onJBIMessage(exchange);
        } else {
            exchange.setError(new JBIException("Bad exchange state : must be Done or Error"));
        }
    }

    /**
     * Process "in-only" MessageExchange
     * 
     * @param exchange
     *            the MessageExchange
     * @throws MessagingException
     *             When errors occur during processing
     */
    private void processInOnlyAsProvider(final Exchange exchange) throws MessagingException {
        this.listener.onJBIMessage(exchange);
        if (exchange.isActiveStatus()) {
            exchange.setDoneStatus();
        }
    }

    /**
     * 
     * @param exchange
     * @throws MessagingException
     */
    private void processInOptionalOutAsConsumer(Exchange exchange) throws MessagingException {
        this.listener.onJBIMessage(exchange);

        // TODO Delete the out message if the it still handled by the message
        // exchange

        // Set Status DONE if no Fault and ACTIVE Status
        if (exchange.getFault() == null && exchange.isActiveStatus()) {
            exchange.setDoneStatus();
        }
    }

    /**
     * Process "in-optional-out" MessageExchange
     * 
     * @param exchange
     *            the MessageExchange
     * @throws MessagingException
     *             When errors occur during processing
     */
    private void processInOptionalOutAsProvider(Exchange exchange) throws MessagingException {
        if (!this.ackFaultReception(exchange)) {
            this.listener.onJBIMessage(exchange);

            // Set Status DONE if no Fault or Out messages and ACTIVE Status
            if (exchange.getFault() == null && !((ExchangeImpl) exchange).isOutMessage()
                    && exchange.isActiveStatus()) {
                exchange.setDoneStatus();
            }
        }
    }

    /**
     * 
     * @param exchange
     * @throws MessagingException
     */
    private void processInOutAsConsumer(Exchange exchange) throws MessagingException {
        this.listener.onJBIMessage(exchange);
        if (exchange.isActiveStatus()) {
            exchange.setDoneStatus();
        }
    }

    /**
     * Process "in-out" MessageExchange
     * 
     * @param exchange
     *            the MessageExchange
     * @throws MessagingException
     *             When errors occur during processing
     * 
     */
    private void processInOutAsProvider(final Exchange exchange, boolean exchangeTerminated)
            throws MessagingException {
        try {
            this.listener.onJBIMessage(exchange);
        } catch (Throwable e) {
            this.logger.log(Level.SEVERE, e.getMessage(), e);
            // set Fault message if an error has been raised
            if (exchange.getFault() == null) {
                exchange.setFault(new SOAP11FaultServerException(e.getMessage(), this
                        .getActorName()));
            }
        }
        // set Fault message if no Fault or Out message
        if (!exchangeTerminated && exchange.getFault() == null
                && !((ExchangeImpl) exchange).isOutMessage()) {
            // an error may have been set as in contrary of the status, it is
            // not forbidden to set an error on the exchange in this transition
            // of this pattern
            if (exchange.getError() != null) {
                exchange.setFault(new SOAP11FaultServerException(exchange.getError().getMessage(),
                        this.getActorName()));
            } else {
                exchange.setFault(new SOAP11FaultServerException(
                        "The Provider do not return any response", this.getActorName()));
            }
        }
    }

    /**
     * 
     * @param exchange
     * @throws MessagingException
     */
    private void processRobustInOnlyAsConsumer(Exchange exchange) throws MessagingException {
        try {
            this.listener.onJBIMessage(exchange);
        } finally {
            if (exchange.isActiveStatus()) {
                exchange.setDoneStatus();
            }
        }
    }

    /**
     * Process "robust-in-only" MessageExchange
     * 
     * @param exchange
     *            the MessageExchange
     * @throws MessagingException
     *             When errors occur during processing
     * 
     */
    private void processRobustInOnlyAsProvider(final Exchange exchange) throws MessagingException {
        this.listener.onJBIMessage(exchange);

        // Set Status DONE if no Fault and ACTIVE Status
        if (exchange.getFault() == null && exchange.isActiveStatus()) {
            exchange.setDoneStatus();
        }
    }

    /**
     * @return The actor name used in Fault in the field &lt;actor&gt;
     */
    protected URI getActorName() {
        return URI.create(this.listener.getProvides().getInterfaceName().getNamespaceURI());
    }

    protected AbstractComponent getComponent() {
        return this.component;
    }

    protected DeliveryChannel getDeliveryChannel() {
        return this.channel;
    }

    protected Transaction getCurrentTransaction() {
        return this.currentTransaction;
    }

    protected AbstractJBIListener getListener() {
        return this.listener;
    }
}
